@@ -27,6 +27,8 @@
 #include "gkm-gnome2-storage.h"
 
 #include "gkm/gkm-certificate.h"
+#define DEBUG_FLAG GKM_DEBUG_STORAGE
+#include "gkm/gkm-debug.h"
 #include "gkm/gkm-data-asn1.h"
 #include "gkm/gkm-manager.h"
 #include "gkm/gkm-module.h"
@@ -34,6 +36,7 @@
 #include "gkm/gkm-serializable.h"
 #include "gkm/gkm-util.h"
 
+#include "egg/dotlock.h"
 #include "egg/egg-asn1x.h"
 #include "egg/egg-asn1-defs.h"
 #include "egg/egg-dn.h"
@@ -84,52 +87,10 @@ struct _GkmGnome2Storage {
 
 G_DEFINE_TYPE (GkmGnome2Storage, gkm_gnome2_storage, GKM_TYPE_STORE);
 
-#define MAX_LOCK_TRIES 20
+#define LOCK_TIMEOUT  4000  /*(4 seconds)*/
 
 #define UNWANTED_IDENTIFIER_CHARS  ":/\\<>|\t\n\r\v "
 
-/* -----------------------------------------------------------------------------
- * HELPERS
- */
-
-#ifndef HAVE_FLOCK
-#define LOCK_SH 1
-#define LOCK_EX 2
-#define LOCK_NB 4
-#define LOCK_UN 8
-
-static int flock(int fd, int operation)
-{
-	struct flock flock;
-
-	switch (operation & ~LOCK_NB) {
-	case LOCK_SH:
-		flock.l_type = F_RDLCK;
-		break;
-	case LOCK_EX:
-		flock.l_type = F_WRLCK;
-		break;
-	case LOCK_UN:
-		flock.l_type = F_UNLCK;
-		break;
-	default:
-		errno = EINVAL;
-		return -1;
-	}
-
-	flock.l_whence = 0;
-	flock.l_start = 0;
-	flock.l_len = 0;
-
-	return fcntl (fd, (operation & LOCK_NB) ? F_SETLK : F_SETLKW, &flock);
-}
-#endif /* !HAVE_FLOCK */
-
-/* -----------------------------------------------------------------------------
- * INTERNAL
- */
-
-
 static gchar*
 name_for_subject (const guchar *subject, gsize n_subject)
 {
@@ -218,12 +179,66 @@ type_from_identifier (const gchar *identifier)
 	return type_from_extension (ext);
 }
 
+static dotlock_t
+lock_and_open_file (const gchar *filename,
+                    gint flags)
+{
+	dotlock_t lockh;
+	gint fd = -1;
+
+	/*
+	 * In this function we don't actually put the object into a 'write' state,
+	 * that's the callers job if necessary.
+	 */
+
+	fd = open (filename, O_RDONLY | O_CREAT, S_IRUSR | S_IWUSR);
+	if (fd == -1) {
+		g_message ("couldn't open store file: %s: %s",
+		           filename, g_strerror (errno));
+		return NULL;
+	}
+
+	lockh = dotlock_create (filename, 0);
+	if (!lockh) {
+		g_message ("couldn't create lock for store file: %s: %s",
+		           filename, g_strerror (errno));
+		close (fd);
+		return NULL;
+	}
+
+	if (dotlock_take (lockh, LOCK_TIMEOUT)) {
+		if (errno == EACCES)
+			g_message ("couldn't write to store file:"
+			           " %s: file is locked", filename);
+		else
+			g_message ("couldn't lock store file: %s: %s",
+			           filename, g_strerror (errno));
+		dotlock_destroy (lockh);
+		close (fd);
+		return NULL;
+	}
+
+	/* Successfully opened file */;
+	dotlock_set_fd (lockh, fd);
+	return lockh;
+}
+
 static gboolean
 complete_lock_file (GkmTransaction *transaction, GObject *object, gpointer data)
 {
-	int fd = GPOINTER_TO_INT (data);
+	GkmGnome2Storage *self = GKM_GNOME2_STORAGE (object);
+	dotlock_t lockh = data;
+	int fd = dotlock_get_fd (lockh);
 
-	/* This also unlocks the file */
+	gkm_debug ("closing: %s", self->filename);
+
+	/*
+	 * Note that there is no error checking for release and
+	 * destroy.  These functions will log errors anyway and we
+	 * can't do much else than logging those errors.
+	 */
+	dotlock_release (lockh);
+	dotlock_destroy (lockh);
 	close (fd);
 
 	/* Completed successfully */
@@ -233,8 +248,7 @@ complete_lock_file (GkmTransaction *transaction, GObject *object, gpointer data)
 static gint
 begin_lock_file (GkmGnome2Storage *self, GkmTransaction *transaction)
 {
-	guint tries = 0;
-	gint fd = -1;
+	dotlock_t lockh;
 
 	/*
 	 * In this function we don't actually put the object into a 'write' state,
@@ -245,42 +259,17 @@ begin_lock_file (GkmGnome2Storage *self, GkmTransaction *transaction)
 	g_assert (GKM_IS_TRANSACTION (transaction));
 
 	g_return_val_if_fail (!gkm_transaction_get_failed (transaction), -1);
+	gkm_debug ("modifying: %s", self->filename);
 
-	/* File lock retry loop */
-	for (tries = 0; TRUE; ++tries) {
-		if (tries > MAX_LOCK_TRIES) {
-			g_message ("couldn't write to store file: %s: file is locked", self->filename);
-			gkm_transaction_fail (transaction, CKR_FUNCTION_FAILED);
-			return -1;
-		}
-
-		fd = open (self->filename, O_RDONLY | O_CREAT, S_IRUSR | S_IWUSR);
-		if (fd == -1) {
-			g_message ("couldn't open store file: %s: %s", self->filename, g_strerror (errno));
-			gkm_transaction_fail (transaction, CKR_FUNCTION_FAILED);
-			return -1;
-		}
-
-		if (flock (fd, LOCK_EX | LOCK_NB) < 0) {
-			if (errno != EWOULDBLOCK) {
-				g_message ("couldn't lock store file: %s: %s", self->filename, g_strerror (errno));
-				close (fd);
-				gkm_transaction_fail (transaction, CKR_FUNCTION_FAILED);
-				return -1;
-			}
-
-			close (fd);
-			fd = -1;
-			g_usleep (200000);
-			continue;
-		}
-
-		/* Successfully opened file */;
-		gkm_transaction_add (transaction, self, complete_lock_file, GINT_TO_POINTER (fd));
-		return fd;
+	lockh = lock_and_open_file (self->filename, O_RDONLY | O_CREAT);
+	if (!lockh) {
+		gkm_transaction_fail (transaction, CKR_FUNCTION_FAILED);
+		return -1;
 	}
 
-	g_assert_not_reached ();
+	/* Successfully opened file */;
+	gkm_transaction_add (transaction, self, complete_lock_file, lockh);
+	return dotlock_get_fd (lockh);
 }
 
 static gboolean
@@ -386,40 +375,42 @@ static gboolean
 begin_modification_state (GkmGnome2Storage *self, GkmTransaction *transaction)
 {
 	GkmDataResult res;
-	struct stat sb;
 	CK_RV rv = CKR_OK;
 
+	/* Already in write state for this transaction? */
+	if (self->transaction != NULL) {
+		g_return_val_if_fail (self->transaction == transaction, FALSE);
+		return TRUE;
+	}
+
 	if (!begin_write_state (self, transaction))
 		return FALSE;
 
 	/* See if file needs updating */
-	if (fstat (self->read_fd, &sb) >= 0 && sb.st_mtime != self->last_mtime) {
+	res = gkm_gnome2_file_read_fd (self->file, self->read_fd, self->login);
+	switch (res) {
+	case GKM_DATA_FAILURE:
+		g_message ("failure updating user store file: %s", self->filename);
+		rv = CKR_FUNCTION_FAILED;
+		break;
+	case GKM_DATA_LOCKED:
+		rv = CKR_USER_NOT_LOGGED_IN;
+		break;
+	case GKM_DATA_UNRECOGNIZED:
+		g_message ("unrecognized or invalid user store file: %s", self->filename);
+		rv = CKR_FUNCTION_FAILED;
+		break;
+	case GKM_DATA_SUCCESS:
+		rv = CKR_OK;
+		break;
+	default:
+		g_assert_not_reached ();
+		break;
+	}
 
-		res = gkm_gnome2_file_read_fd (self->file, self->read_fd, self->login);
-		switch (res) {
-		case GKM_DATA_FAILURE:
-			g_message ("failure updating user store file: %s", self->filename);
-			rv = CKR_FUNCTION_FAILED;
-			break;
-		case GKM_DATA_LOCKED:
-			rv = CKR_USER_NOT_LOGGED_IN;
-			break;
-		case GKM_DATA_UNRECOGNIZED:
-			g_message ("unrecognized or invalid user store file: %s", self->filename);
-			rv = CKR_FUNCTION_FAILED;
-			break;
-		case GKM_DATA_SUCCESS:
-			rv = CKR_OK;
-			break;
-		default:
-			g_assert_not_reached ();
-			break;
-		}
-
-		if (rv != CKR_OK) {
-			gkm_transaction_fail (transaction, rv);
-			return FALSE;
-		}
+	if (rv != CKR_OK) {
+		gkm_transaction_fail (transaction, rv);
+		return FALSE;
 	}
 
 	/* Write out the data once completed with modifications */
@@ -708,15 +699,17 @@ static CK_RV
 refresh_with_login (GkmGnome2Storage *self, GkmSecret *login)
 {
 	GkmDataResult res;
+	dotlock_t lockh;
 	struct stat sb;
 	CK_RV rv;
 	int fd;
 
 	g_assert (GKM_GNOME2_STORAGE (self));
+	gkm_debug ("refreshing: %s", self->filename);
 
 	/* Open the file for reading */
-	fd = open (self->filename, O_RDONLY, 0);
-	if (fd == -1) {
+	lockh = lock_and_open_file (self->filename, O_RDONLY);
+	if (!lockh) {
 		/* No file, no worries */
 		if (errno == ENOENT)
 			return login ? CKR_USER_PIN_NOT_INITIALIZED : CKR_OK;
@@ -724,6 +717,8 @@ refresh_with_login (GkmGnome2Storage *self, GkmSecret *login)
 		return CKR_FUNCTION_FAILED;
 	}
 
+	fd = dotlock_get_fd (lockh);
+
 	/* Try and update the last read time */
 	if (fstat (fd, &sb) >= 0)
 		self->last_mtime = sb.st_mtime;
@@ -753,7 +748,17 @@ refresh_with_login (GkmGnome2Storage *self, GkmSecret *login)
 	if (rv == CKR_FUNCTION_FAILED)
 		self->last_mtime = 0;
 
+	gkm_debug ("closing: %s", self->filename);
+
+	/*
+	 * Note that there is no error checking for release and
+	 * destroy.  These functions will log errors anyway and we
+	 * can't do much else than logging those errors.
+	 */
+	dotlock_release (lockh);
+	dotlock_destroy (lockh);
 	close (fd);
+
 	return rv;
 }
 
@@ -775,16 +780,16 @@ gkm_gnome2_storage_real_read_value (GkmStore *base, GkmObject *object, CK_ATTRIB
 	g_return_val_if_fail (GKM_IS_OBJECT (object), CKR_GENERAL_ERROR);
 	g_return_val_if_fail (attr, CKR_GENERAL_ERROR);
 
-	identifier = g_hash_table_lookup (self->object_to_identifier, object);
-	if (!identifier)
-		return CKR_ATTRIBUTE_TYPE_INVALID;
-
 	if (self->last_mtime == 0) {
 		rv = gkm_gnome2_storage_refresh (self);
 		if (rv != CKR_OK)
 			return rv;
 	}
 
+	identifier = g_hash_table_lookup (self->object_to_identifier, object);
+	if (!identifier)
+		return CKR_ATTRIBUTE_TYPE_INVALID;
+
 	res = gkm_gnome2_file_read_value (self->file, identifier, attr->type, &value, &n_value);
 	switch (res) {
 	case GKM_DATA_FAILURE:
@@ -817,20 +822,15 @@ gkm_gnome2_storage_real_write_value (GkmStore *base, GkmTransaction *transaction
 	g_return_if_fail (!gkm_transaction_get_failed (transaction));
 	g_return_if_fail (attr);
 
+	if (!begin_modification_state (self, transaction))
+		return;
+
 	identifier = g_hash_table_lookup (self->object_to_identifier, object);
 	if (!identifier) {
 		gkm_transaction_fail (transaction, CKR_ATTRIBUTE_READ_ONLY);
 		return;
 	}
 
-	if (self->last_mtime == 0) {
-		rv = gkm_gnome2_storage_refresh (self);
-		if (rv != CKR_OK) {
-			gkm_transaction_fail (transaction, rv);
-			return;
-		}
-	}
-
 	res = gkm_gnome2_file_write_value (self->file, identifier, attr->type, attr->pValue, attr->ulValueLen);
 	switch (res) {
 	case GKM_DATA_FAILURE:
